#include "massiveDIS.hh"
#include <gsl/gsl_sf.h>
#include <gsl/gsl_errno.h>
#include "integration.hh"
#include <stdlib.h>
#include <iostream>
#include "math.hh"

using namespace std;

double CF = 4./3;

/*
	Constructor
 */
MassiveDIS::MassiveDIS(const process p, const parton part, std::string PDFset, int member):
fProcess(p),
fParton(part),
fMuRQ(1),
fMuFQ(1),
fM1(0.01), // GeV
fM2(0.01), // GeV
fPrec(1e-3)
{
	fPDF = new PDFinterface::PDFs(PDFset, member);
	fCounting = standard;
	fDamping = false;
	double m = fPDF->getQuarkMass(part);
	if (abs(part) < 4) m = 0.01; 
	if(m>0) fM1 = m;
	if(p<0)
	{
		fM2 = fPDF->getQuarkMass(CCPartner(part));
		if (abs(CCPartner(part)) < 4) fM2 = 0.01;
	}
	if(p>0) fM2 = m;
}

/*
	Destructor
 */
MassiveDIS::~MassiveDIS()
{
  delete fPDF;
}

void MassiveDIS::ReInitPDF(std::string PDFset, int member) {
  delete fPDF;
  fPDF = new PDFinterface::PDFs(PDFset, member);
}


/*
	
 */
double MassiveDIS::F(const double x, const double Q, order ord, scheme sc, sfunction fun, const double muF_Q_ratio, const double muR_Q_ratio)
{
  fMuRQ = muR_Q_ratio;
  fMuFQ = muF_Q_ratio;
  if(Q*fMuFQ<fM1 && sc!=FO) return F(x, Q, ord, FO, fun);
  order gluonord = (fCounting == primed ? (order) ((int) ord + 1) : ord );
  double damp = (fDamping == false ? 1 : pow((1.-pow(fM1/Q,2)),2)  );
  double result = 0;
  InitCouplings(fProcess,fParton,Q,fun);

  switch(sc)
    {
    case FO:
      result = F_g(x,Q,gluonord,fun);
      result += F_q(x,Q,ord,fun); // this should be zero if IC=0 and the PDFset is a FFS (3FS) one
      break;
      
    case FONLL:
      result = F_g(x,Q,gluonord,fun) + damp*(F_q0(x,Q,ord,fun)-F_g_sub0(x,Q,gluonord,fun));
      break;
      
    case ACOT:
      result = F_g(x,Q,gluonord,fun) -F_g_sub(x,Q,gluonord,fun) + F_q(x,Q,ord,fun) -F_q_sub(x,Q,ord,fun);
      break;
      
    case ZM:
      result = F_q0(x,Q,ord,fun) + F_g0(x,Q,gluonord,fun);
      break;

    case FOQ:
      result =  F_q(x,Q,ord,fun);
      break;
      
    default:
      cout << "Warning: scheme not available. Abort." << endl;
      abort();
    }

  //now add contribution for the second quark
  
  //exchange the masses
  double fm1temp = fM1;
  fM1 = fM2;
  fM2 = fm1temp;
  //switch parton
  parton fpartontemp = fParton;
  fParton = (fProcess == DISCC ? CCPartner(fParton) : (parton) -fParton);
  
  switch(sc)
    {
    case FO:
      result += (fun == F_3 ? -1. : 1. )*F_q(x,Q,ord,fun); 
      break;
      
    case FONLL:
      result += (fun == F_3 ? -1. : 1. )*damp*(F_q0(x,Q,ord,fun));
      break;
      
    case ACOT:
      result += (fun == F_3 ? -1. : 1. )*(F_q(x,Q,ord,fun) -F_q_sub(x,Q,ord,fun));
      break;
      
    case ZM:
      result += (fun == F_3 ? -1. : 1. )*F_q0(x,Q,ord,fun);
      break;

    case FOQ:
      result += (fun == F_3 ? -1. : 1. )*F_q(x,Q,ord,fun);
      break;
      
    default:
      cout << "Warning: scheme not available. Abort." << endl;
      abort();
    }

    // TR counting
    if(fCounting == TRcounting && sc!=FO && sc!=ZM) {
    	if(Q*fMuFQ<fM1) return result = F(x, Q, (order) ((int) gluonord + 1), FO, fun);
    	result += F(x, fM1, (order) ((int) gluonord + 1), FO, fun);
    }
  
    //exchange the masses again
    fm1temp = fM1;
    fM1 = fM2;
    fM2 = fm1temp;
    //switch parton
    fParton = fpartontemp;

    return result;
}

double MassiveDIS::F_g(const double x, const double Q, order ord, sfunction fun)
{
	if (fun == F_L) return F_g(x,Q,ord,F_2)-2.*x*F_g(x,Q,ord,F_1);
	double result = 0;
	double pref = 1.;
	if(ord > 0)
	{
		result += fPDF->alphasQ(Q*fMuRQ)/M_PI*Fgluon(x,Q,Q*fMuFQ,fun);
	}
	return pref*result;
}

double MassiveDIS::F_g_sub (const double x, const double Q, order ord, sfunction fun)
{
	if (fun == F_L) return F_g_sub(x,Q,ord,F_2)-2.*x*F_g_sub(x,Q,ord,F_1);
	if(Q*fMuFQ<fM1) return 0;
	double result = 0;
	double pref = prefactor_massive(x,Q,fun);
	if(ord > 0)
	{
		result += fPDF->alphasQ(Q*fMuRQ)/M_PI*g_sub(chi(x,Q),Q*fMuFQ,Q*fMuFQ,fun);
	}
	return pref*result;	
}

double MassiveDIS::F_g0(const double x, const double Q, order ord, sfunction fun)
{
	if (fun == F_L) return F_g0(x,Q,ord,F_2)-2.*x*F_g0(x,Q,ord,F_1);
	double result = 0;
	double pref = prefactor_massless(x,fun);
	if(ord > 0)
	{
		result += fPDF->alphasQ(Q*fMuRQ)/M_PI*Fgluon0(x,Q,Q*fMuFQ,fun);
	}
	return pref*result;
}

double MassiveDIS::F_g_sub0(const double x, const double Q, order ord, sfunction fun)
{
	if (fun == F_L) return F_g_sub0(x,Q,ord,F_2)-2.*x*F_g_sub0(x,Q,ord,F_1);
	if(Q*fMuFQ<fM1) return 0;
	double result = 0;
	double pref = prefactor_massless(x,fun);
	if(ord > 0)
	{
		result += fPDF->alphasQ(Q*fMuRQ)/M_PI*g_sub(x,Q*fMuFQ,Q*fMuFQ,fun);
	}
	return pref*result;	
}

double MassiveDIS::F_q(const double x, const double Q, order ord, sfunction fun)
{
	if (fun == F_L) return F_q(x,Q,ord,F_2)-2.*x*F_q(x,Q,ord,F_1);
	double pref1 = prefactor_massive(x,Q,fun);	
	double result1 = fPDF->xfx(chi(x,Q), Q*fMuFQ, fParton)/chi(x,Q);
	if(ord > 0)
	{
		result1 += CF*fPDF->alphasQ(Q*fMuRQ)/2./M_PI*F_q_as(x,Q,Q*fMuFQ,fun, fParton);
	}
	return pref1*result1;
}

double MassiveDIS::F_q_sub (const double x, const double Q, order ord, sfunction fun)
{
	if (fun == F_L) return F_q_sub(x,Q,ord,F_2)-2.*x*F_q_sub(x,Q,ord,F_1);
	double pref1 = prefactor_massive(x,Q,fun);
	double result1 = 0;
	if(ord > 0)
	{
		result1 += CF*fPDF->alphasQ(Q*fMuRQ)/2./M_PI*q_sub(chi(x,Q),Q*fMuFQ,Q*fMuFQ,fun, fParton);
	}
	return pref1*result1;
}

double MassiveDIS::F_q0    (const double x, const double Q, order ord, sfunction fun)
{
	if (fun == F_L) return F_q0(x,Q,ord,F_2)-2.*x*F_q0(x,Q,ord,F_1);
	double pref1 = prefactor_massless(x,fun);
	double result1 = fPDF->xfx(x, Q*fMuFQ, fParton)/x;
	if(ord > 0)
	{
		result1 += CF*fPDF->alphasQ(Q*fMuRQ)/2./M_PI*F_q_as_0(x,Q,Q*fMuFQ,fun, fParton);
	}
	return pref1*result1;
}

double MassiveDIS::F_q_sub0(const double x, const double Q, order ord, sfunction fun)
{
	if (fun == F_L) return F_q_sub0(x,Q,ord,F_2)-2.*x*F_q_sub0(x,Q,ord,F_1);
	double pref1 = prefactor_massless(x,fun);
	double result1 = 0;
	if(ord > 0)
	{
		result1 += CF*fPDF->alphasQ(Q*fMuRQ)/2./M_PI*q_sub(x,Q*fMuFQ,Q*fMuFQ,fun,fParton);
	}
	return pref1*result1;
}

double MassiveDIS::prefactor_massive (const double x, const double Q, sfunction fun)
{
	switch(fun)
	{
		case F_1:
			return F1pref(Q);
			break;
		case F_2:
			return F2pref(x,Q);
			break;
		case F_3:
			return F3pref(Q);
			break;

		default:
			cout << "Warning: F_i not available. Abort." << endl;
			abort();
	}

}

double MassiveDIS::prefactor_massless(const double x, sfunction fun)
{
	switch(fun)
	{
		case F_1:
			return F1masslesspref();
			break;
		case F_2:
			return F2masslesspref(x);
			break;
		case F_3:
			return F3masslesspref();
			break;

		default:
			cout << "Warning: F_i not available. Abort." << endl;
			abort();
	}
}

/*
	massive quark
 */

double MassiveDIS::F_q_as(const double x, const double Q, const double muF, sfunction fun, parton part)
{
	double delta, result;
	delta = DeltaTerms(Q,fun);
	result = (fPDF->xfx(chi(x,Q), muF, part)/chi(x,Q))*delta;
	result += PlusPrescription(x, Q, muF, fun, part); 
	return result;
}


void MassiveDIS::InitCouplings(const process p, const parton part, const double Q, const sfunction fun)
{
  fSp = fSm = fRp = fRm = fQp = fQm = 0;
  bool run[4] = {0,0,0,0};  // gamma, Z, Z-gamma interference, W
  switch(p)
    {
    case DISNC:
      run[0] = run[1] = run[2] = 1;
      break;
    case DISNCGAMMA:
      run[0] = 1;
      break;
    case DISNCZ:
      run[1] = 1;
      break;
    case DISNCZGAMMA:
      run[2] = 1;
      break;
    case DISCC:
      run[3] = 1;
      break;
    default:
      cout << "Warning: process not available. Abort." << endl;
      abort();
    }
  double sin2W = 0.23126; //MSbar scheme
  double ptype = (abs(part)%2==0 ? 1. : -1.);
  double eq = (abs(part)%2==0 ? 2./3. : -1./3.);
  double a = (part>0 ? 1.: -1.);
  //double V [4] = {eq, ptype/2-a*2.*eq*sin2W, ptype/2-a*2.*eq*sin2W, 1};
  //double Vp[4] = {eq, ptype/2-a*2.*eq*sin2W, eq,                    1};
  //cout << "a" << a << endl;
  //cout << "ptype" << a << endl;
  double V [4] = {eq, ptype/2-a*2.*eq*sin2W, ptype/2-a*2.*eq*sin2W, 1};
  double Vp[4] = {eq, ptype/2-a*2.*eq*sin2W, eq,                    1};
  double A [4] = { 0, ptype/2., ptype/2., 1};
  double Ap[4] = { 0, ptype/2., 0,        1};
  double w[4] = {1,1,2,1};
  if(p==DISNC) {
    //double GF = 1.1663787e-5;
    double MZ = 91.1876;
    //double alpha = 1./137.036;
    //double etaZ = GF*MZ*MZ/2./sqrt(2.)/M_PI/alpha*Q*Q/(Q*Q+MZ*MZ);
    double etaZ = Q*Q / ( Q*Q + MZ*MZ )/ (4 * sin2W * ( 1 - sin2W ) );  //APFEL etaZ
    double echarge = -1; // 1 for positron, -1 for electron; ineffective if lambda=0
    double lambda = 0; // helicity of the e+-; assuming we average over lambda, it is effectively zero
    double gVe = -1./2+2.*sin2W;
    double gAe = -1./2.;
    if(fun == F_3) {
      w[0] = 0;
      w[1] *= (2*gVe*gAe+echarge*lambda*(gVe*gVe+gAe*gAe))*etaZ*etaZ;
      w[2] *= -(gAe+echarge*lambda*gVe)*etaZ;
    } else {
      w[0] *= 1.; 
      w[1] *= (gVe*gVe+gAe*gAe+echarge*2*lambda*gVe*gAe)*etaZ*etaZ;
      w[2] *= -(gVe+echarge*lambda*gAe)*etaZ;
    }
  }
  for(int i=0; i<4; i++) {
    if(run[i]) {
      fSp += w[i] * (V[i]*Vp[i] + A [i]*Ap[i]);
      fSm += w[i] * (V[i]*Vp[i] - A [i]*Ap[i]);
      fRp += w[i] * (V[i]*Ap[i] + Vp[i]*A [i])/2.;
      //fRm += w[i] * (V[i]*Ap[i] - Vp[i]*A [i])/2.; // this is non-zero only for Z-gamma interference, but once gamma-Z is added it cancels
      //fQp += w[i] * (V[i]*V [i] + A [i]*A [i]);
      //fQm += w[i] * (V[i]*V [i] - A [i]*A [i]);
    }
  }
  fQp = fSp;
  fQm = fSm;
}

/***********************************************************************
******************************DELTA IC**********************************
***********************************************************************/

double MassiveDIS::DeltaF2IC(const double x, const double Q, const double mu0, order ord)
{
	if(mu0<fM1) return 0;
	double as = fPDF->alphasQ(Q*fMuRQ);
	double pref1, prefmassless1, result1, pref2, prefmassless2, result2;
	InitCouplings(fProcess,fParton,Q,F_2);
	pref1 = F2pref(x,Q);
	prefmassless1 = F2masslesspref(x);
	result1 = pref1*(fPDF->xfx(chi(x,Q), mu0, fParton))/chi(x,Q)-prefmassless1*(fPDF->xfx(x, mu0, fParton))/x;
	if(ord > 0)
	{
		double resultNLO=0;
		resultNLO += CF*pref1       *F_q_as(x,Q,mu0,F_2,fParton); 
		resultNLO -= CF*prefmassless1*F_q_as_0(x,mu0,mu0,F_2,fParton); 
		resultNLO -= CF*prefmassless1*q_sub(x,Q,mu0,F_2,fParton);
		resultNLO -= CF*pref1        *q_sub(chi(x,Q),mu0,mu0,F_2,fParton);
		resultNLO += CF*prefmassless1*q_sub(x,mu0,mu0,F_2,fParton);
		resultNLO += 2*prefmassless1 *g_sub(x,mu0,mu0,F_2);        //2 takes into account q and q~
		resultNLO -= 2*pref1         *g_sub(chi(x,Q),mu0,mu0,F_2); //2 takes into account q and q~
		resultNLO *= as/2./M_PI;
		result1 += resultNLO;
	}
	pref2 = F2pref(x,Q);
	prefmassless2 = F2masslesspref(x);
	result2 = pref2*(fPDF->xfx(chi(x,Q), mu0, -fParton))/chi(x,Q)-prefmassless2*(fPDF->xfx(x, mu0, -fParton))/x;
	if(ord > 0)
	{
		double resultNLO=0;
		resultNLO += CF*pref2       *F_q_as(x,Q,mu0,F_2,(parton) -fParton); 
		resultNLO -= CF*prefmassless2*F_q_as_0(x,mu0,mu0,F_2,(parton) -fParton); 
		resultNLO -= CF*prefmassless2*q_sub(x,Q,mu0,F_2,(parton) -fParton);
		resultNLO -= CF*pref2        *q_sub(chi(x,Q),mu0,mu0,F_2,(parton) -fParton);
		resultNLO += CF*prefmassless2*q_sub(x,mu0,mu0,F_2,(parton) -fParton);
		resultNLO *= as/2./M_PI;
		result2 += resultNLO;
	}
	return result1+result2;
}


/*
	Auxiliary functions
 */

//triangle function
double MassiveDIS::DeltaFun(const double a, const double b, const double c)
{
	double result;
	result = sqrt(pow(a,2)+pow(b,2)+pow(c,2)-2.*(a*b+b*c+c*a));
	return result;
}

double MassiveDIS::Delta(const double Q)
{
	double result;
	result = DeltaFun(pow(fM1,2),pow(fM2,2),-pow(Q,2));
	return result;
}

double MassiveDIS::Deltaprime(double xi, const double Q)
{
	double result;
	result = DeltaFun(pow(fM1,2),shat(xi,Q),-pow(Q,2));
	return result;
}

//scaling variable chi
double MassiveDIS::chi(double x, const double Q)
{
	double result;
	result = x/2./pow(Q,2)*(Sigmapm(Q)+Delta(Q));
	return result;
}

double D0(double z)
{
	return 1./(1.-z);
}

double Plus_g(double z, const double Q, MassiveDIS *MDIS, sfunction fun)
{
	double result;
	switch(fun)
	{
		case F_1:
		result = MDIS->GetPlusF1(z,Q);
		break;

		case F_2:
		result = MDIS->GetPlusF2(z,Q);
		break;

		case F_3:
		result = MDIS->GetPlusF3(z,Q);
		break;

		default:
			cout << "Error. Abort." << endl;
			abort();
	}
	return result;
}


struct s_plus  {double Q, muF, chi; sfunction fun; MassiveDIS *MDIS; PDFinterface::PDFs *pdf; parton p;};
double Plus_integrand(double t, void* par) 
{
	s_plus pars = *(s_plus *) par;
    PDFinterface::PDFs *pdf  = pars.pdf;
    MassiveDIS *MDIS = pars.MDIS;
    double muF = pars.muF;
    double Q = pars.Q;
    double chi = pars.chi;
    sfunction fun = pars.fun;
    parton p = pars.p;
    double yt = chi+(1.-chi)*t;
    double xt = chi/yt;
    double Fx  = pdf->xfx(chi,muF,p)/chi;
    double Fxt = pdf->xfx(xt,muF,p)/xt;
    //the plus prescription is defined between chi and 1
    //int_chi^1 C Q_1 dy/y
    //Q_1 = pdf, C = g[f]_+, f = 1/(1-z)_+, g = (1-z)*(shatp)/(8*shat)*1/N*fhat
    double result;
    result = (1.-chi)*D0(yt)*Plus_g(yt,Q,MDIS,fun)*(1./(yt)*Fxt-Fx)
    +Fx*(-chi*D0(chi*t)*Plus_g(chi*t,Q,MDIS,fun)+(Plus_g(t,Q,MDIS,fun)-Plus_g(1,Q,MDIS,fun))*D0(t));
    return result;
}

double MassiveDIS::PlusPrescription(double x, const double Q, double muF, sfunction fun, parton part) 
{
	s_plus pars = {Q, muF, chi(x,Q), fun, this, fPDF, part};
	double res=0;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    res = integration::gauss(Plus_integrand,0.,1.,fPrec,NULL,&pars);
    gsl_set_error_handler (old_handler);
    return res;
}

double MassiveDIS::DeltaTerms(const double Q, const sfunction fun)
{
	double result;
	switch(fun)
	{
		case F_1:
			result = S1_nopoles(Q)+V1_nopoles(Q);
			break;

		case F_2:
			result = S2_nopoles(Q)+V2_nopoles(Q);
			break;

		case F_3:
			result = S3_nopoles(Q)+V3_nopoles(Q);
			break;

		default:
			cout << "Error. Abort." << endl;
			abort();
	}
	return result;
}


/*********************Delta terms********************/

//I1
double MassiveDIS::I1(const double Q)
{
	double Del, Spp, result;
	Del = Delta(Q);
	Spp = Sigmapp(Q);
	result = 1./Del*log((Spp+Del)/(Spp-Del));
	return result;
}

double MassiveDIS::Cp(const double Q)
{
	double result;
	result = (2.*fM1*fM2*I1(Q));
	return result;
}


double MassiveDIS::C1p(const double Q)
{
	double result;
	result = -1./pow(Q,2)*(Sigmapm(Q)*I1(Q)+log(pow(fM1,2)/pow(fM2,2))); 
	return result;
}

double MassiveDIS::C1m(const double Q)
{
	double result;
	result = -1./pow(Q,2)*(Sigmamp(Q)*I1(Q)-log(pow(fM1,2)/pow(fM2,2)));
	return result;
}

double MassiveDIS::Cqm(const double Q)
{
	double result;
	result = 1./pow(Q,4)*(
		(pow(Delta(Q),2))-2.*pow(fM1,2)*pow(Q,2))*I1(Q)-2.*pow(Q,2)
	+Sigmapm(Q)*log(pow(fM1,2)/pow(fM2,2));
	return result;
}

double MassiveDIS::Cqp(const double Q)
{
	double result;
	result = 1./pow(Q,4)*(
		(-pow(Delta(Q),2))+2.*pow(fM2,2)*pow(Q,2)-Sigmamp(Q)*pow(Q,2))*I1(Q)+2.*pow(Q,2)
	+(Sigmamp(Q)+pow(Q,2))*log(pow(fM1,2)/pow(fM2,2));
	return result;
}

double MassiveDIS::CRm_nopoles(const double Q)
{
	double Del, Spp, Smp, Spm, result;
	Del = Delta(Q);
	Spp = Sigmapp(Q);
	Spm = Sigmapm(Q);
	Smp = Sigmamp(Q);
	result = (pow(Del,2)/2./pow(Q,2)+Spp*(1.+log(pow(Q,2)/Del)))*I1(Q)
	+(pow(fM2,2)-pow(fM1,2))/2./pow(Q,2)*log(pow(fM1,2)/pow(fM2,2))-log(pow(Q,2)/pow(fM1,2))-log(pow(Q,2)/pow(fM2,2))-4.
	+Spp/Del*(1./2.*pow(log(fabs((Del-Spm)/(2*pow(Q,2)))),2)+1./2.*pow(log(fabs((Del-Smp)/(2*pow(Q,2)))),2)-1./2.*pow(log(fabs((Del+Spm)/(2*pow(Q,2)))),2)-1./2.*pow(log(fabs((Del+Smp)/(2*pow(Q,2)))),2)
		-Li2((Del-Spm)/2./Del)-Li2((Del-Smp)/2./Del)+Li2((Del+Spm)/2./Del)+Li2((Del+Smp)/2./Del)
		);
	return result;
}

//soft real contributions
double MassiveDIS::S1_nopoles(const double Q)
{
	double Del, Spp, result;
	Del = Delta(Q);
	Spp = Sigmapp(Q);
	result = 2.+Spp/Del*(Del*I1(Q)+Li2(2.*Del/(Del-Spp))-Li2(2.*Del/(Del+Spp)))
	+log(pow(Del,2)/pow(fM2,2)/pow(Q,2))*(-2.+Spp*I1(Q));
	return result;
}

double MassiveDIS::S2_nopoles(const double Q)
{
	return S1_nopoles(Q);
}

double MassiveDIS::S3_nopoles(const double Q)
{
	return S1_nopoles(Q);
}

//virtual contributions
double MassiveDIS::V1_nopoles(const double Q)
{
	double Spp, result;
	Spp = Sigmapp(Q);
	result = CRm_nopoles(Q) +(fSm*Spp-2.*fSp*fM1*fM2)/(fSp*Spp-2.*fSm*fM1*fM2)*Cp(Q);
	return result;
}

double MassiveDIS::V2_nopoles(const double Q)
{
	double result;
	result = CRm_nopoles(Q)+1./2.*(pow(fM1,2)*C1p(Q)+pow(fM2,2)*C1m(Q))+fSm/fSp*(Cp(Q)+fM1*fM2/2.*(C1p(Q)+C1m(Q)));
	return result;
}


double MassiveDIS::V3_nopoles(const double Q)
{
	double result;
	result = CRm_nopoles(Q)+fRm/fRp*Cp(Q);
	return result;
}

//s1hat
double MassiveDIS::s1hat(double xi, const double Q)
{
	double Del, Spm, result;
	Del = Delta(Q);
	Spm = Sigmapm(Q);
	result = (1.-xi)/(2.*xi)*((Del-Spm)*xi+Del+Spm);
	return result;
}

//shat
double MassiveDIS::shat(double xi, const double Q)
{
	return (s1hat(xi,Q) + pow(fM2,2));
}

/****************Plus Distribution terms*************/

double MassiveDIS::Lxi(double xi, const double Q)
{
	return log((Sigmapp(Q)+s1hat(xi,Q)-Deltaprime(xi,Q))/(Sigmapp(Q)+s1hat(xi,Q)+Deltaprime(xi,Q)));
}

double MassiveDIS::Ixi(double xi, const double Q)
{
	double s1h;
	s1h = s1hat(xi,Q);
	return ((s1h+2.*pow(fM2,2))/pow(s1h,2)+(s1h+pow(fM2,2))/Deltaprime(xi,Q)/pow(s1h,2)*Sigmapp(Q)*Lxi(xi,Q));
}

double MassiveDIS::IxiReg(double xi, const double Q)
{
	double s1h;
	s1h = s1hat(xi,Q);
	return ((s1h+2.*pow(fM2,2))+(s1h+pow(fM2,2))/Deltaprime(xi,Q)*Sigmapp(Q)*Lxi(xi,Q));
}


//normalisation factors
double MassiveDIS::N1(const double Q)
{
	double result;
	result = (fSp*Sigmapp(Q)-2.*fM1*fM2*fSm)/2./Delta(Q);
	return result;
}

double MassiveDIS::N2(double xi, const double Q)
{
	double result;
	result = 2.*fSp*Delta(Q)/pow(Deltaprime(xi, Q),2);
	return result;
}

double MassiveDIS::N3(double xi, const double Q)
{
	double result;
	result = 2.*fRp/Deltaprime(xi, Q);
	return result;
}

//massive matrix elements
double MassiveDIS::fhat1(double xi, const double Q)
{
	double Del, Delp, Spp, Spm, Smp, s1h, result;
	s1h = s1hat(xi,Q);
	Delp = Deltaprime(xi,Q);
	Del = Delta(Q);
	Spp = Sigmapp(Q);
	Spm = Sigmapm(Q);
	Smp = Sigmamp(Q);
	result = 8./pow(Delp,2)*(-pow(Del,2)*(fSp*Spp-2.*fM1*fM2*fSm)*Ixi(xi,Q)+2.*fM1*fM2*fSm*(1./s1h*(pow(Delp,2)+4.*pow(fM2,2)*Spm)
		+2.*Spm-Smp+(Spp+s1h)/2.+(s1h+pow(fM2,2))/Delp/s1h*(pow(Delp,2)+2.*Spm*Spp+(pow(fM2,2)+pow(Q,2))*s1h)*Lxi(xi,Q))
	+fSp*((-pow(fM2,2)*Spp)/((s1h+pow(fM2,2))*s1h)*(pow(Del,2)+4.*pow(fM2,2)*Spm)-1./4./(s1h+pow(fM2,2))*(3.*pow(Spp,2)*Smp+4.*pow(fM2,2)*(10.*Spp*Spm-Spm*Smp-pow(fM1,2)*Spp)
		+s1h*(-7.*Spp*Smp+18.*pow(Del,2)-4.*pow(fM1,2)*(7.*pow(Q,2)-4*pow(fM2,2)+7.*pow(fM1,2)))
			+3.*pow(s1h,2)*(Spm-2.*pow(fM1,2))-pow(s1h,3))+
		(s1h+pow(fM2,2))/2./Delp*(-2./s1h*Spp*(pow(Del,2)+2.*Spm*Spp)+(4.*pow(fM1,2)*pow(fM2,2)-7.*Spm*Spp)-4.*Spm*s1h-pow(s1h,2))*Lxi(xi,Q))
	);
	return result;
}

double MassiveDIS::fhat2(double xi, const double Q)
{
	double Del, Delp, Spm, Spp, Smp, s1h, result;
	s1h = s1hat(xi,Q);
	Delp = Deltaprime(xi,Q);
	Del = Delta(Q);
	Spp = Sigmapp(Q);
	Spm = Sigmapm(Q);
	Smp = Sigmamp(Q);
	result = 16./pow(Delp,4)*(-2.*pow(Del,4)*fSp*Ixi(xi,Q)+2.*fM1*fM2*fSm*(((s1h+pow(fM2,2))/Delp)*(pow(Delp,2)-6.*pow(fM1,2)*pow(Q,2))*Lxi(xi,Q)
		-pow(Delp,2)*(s1h+Spp)/2./(s1h+pow(fM2,2))+(2.*pow(Delp,2)-3.*pow(Q,2)*(s1h+Spp)))+fSp*(-2.*(pow(Del,2)-6.*pow(fM1,2)*pow(Q,2))*(s1h+pow(fM2,2))
		-2.*(pow(fM1,2)+pow(fM2,2))*pow(s1h,2)-9.*pow(fM2,2)*pow(Spm,2)+pow(Del,2)*(2*Spp-pow(fM2,2))+2.*s1h*(2.*pow(Del,2)+(pow(fM1,2)-5.*pow(fM2,2))*Spm)
			+(pow(Delp,2)-6.*pow(Q,2)*(pow(fM2,2)+s1h))*Spp*(s1h+Spp)/2./(s1h+pow(fM2,2))-2.*pow(Del,2)/s1h*(pow(Del,2)+2.*(2.*pow(fM2,2)+s1h)*Spm)
			+(s1h+pow(fM2,2))/Delp*(-2./s1h*pow(Del,2)*(pow(Del,2)+2.*Spm*Spp)-2.*s1h*(pow(Del,2)-6.*pow(fM1,2)*pow(Q,2))
				-(pow(Delp,2)-18.*pow(fM1,2)*pow(Q,2))*Spp-2.*pow(Del,2)*(Spp+2.*Spm))*Lxi(xi,Q)));
	return result;
}

double MassiveDIS::fhat3(double xi, const double Q)
{
	double Del, Delp, Spm, Spp, Smp, s1h, result;
	s1h = s1hat(xi,Q);
	Delp = Deltaprime(xi,Q);
	Del = Delta(Q);
	Spp = Sigmapp(Q);
	Spm = Sigmapm(Q);
	Smp = Sigmamp(Q);
	result = 16./pow(Delp,2)*(-2.*pow(Del,2)*fRp*Ixi(xi,Q)+2.*fM1*fM2*fRm*(1.-Smp/s1h+(s1h+pow(fM2,2))*(s1h+Spm)/Delp/s1h*Lxi(xi,Q))
		+fRp*(Smp-3.*Spm-2./s1h*(pow(Del,2)+2.*pow(fM2,2)*Spm)-(s1h-Smp)*(s1h+Spp)/2./(s1h+pow(fM2,2))
			+(s1h+pow(fM2,2))/Delp/s1h*(-pow(s1h,2)+4.*(pow(fM1,2)*Smp-pow(Del,2))-3.*s1h*Spm)*Lxi(xi,Q)));
	return result;
}

//xi->1

//s1hatReg
double MassiveDIS::s1hatReg(double xi, const double Q)
{
	double Del, Spm, result;
	Del = Delta(Q);
	Spm = Sigmapm(Q);
	result = 1./(2.*xi)*((Del-Spm)*xi+Del+Spm);
	return result;
}

//massive matrix elements
double MassiveDIS::fhat1Reg(double xi, const double Q)
{
	double Del, Delp, Spp, result;
	Delp = Deltaprime(xi,Q);
	Del = Delta(Q);
	Spp = Sigmapp(Q);
	result = 8./pow(Delp,2)*(-pow(Del,2)*(fSp*Spp-2.*fM1*fM2*fSm)*IxiReg(xi,Q));
	return result;
}

double MassiveDIS::fhat2Reg(double xi, const double Q)
{
	double Del, Delp, result;
	Delp = Deltaprime(xi,Q);
	Del = Delta(Q);
	result = 16./pow(Delp,4)*(-2.*pow(Del,4)*fSp*IxiReg(xi,Q));
	return result;
}

double MassiveDIS::fhat3Reg(double xi, const double Q)
{
	double Del, Delp, result;
	Delp = Deltaprime(xi,Q);
	Del = Delta(Q);
	result = 16./pow(Delp,2)*(-2.*pow(Del,2)*fRp*IxiReg(xi,Q));
	return result;
}


//subtraction terms
	
double f_sub(double t, const double m1, const double mu)
{
	return (1.+pow(t,2))/(1.-t)*(log(pow(mu,2)/pow(m1,2))-1.-2.*log(1-t));
}

struct sub_plus  {double muLog, muF, chi; MassiveDIS *MDIS; PDFinterface::PDFs *pdf; parton p;};
double Sub_integrand(double t, void* par) 
{
	sub_plus pars = *(sub_plus *) par;
    PDFinterface::PDFs *pdf  = pars.pdf;
    MassiveDIS *MDIS = pars.MDIS;
    double Q = pars.muLog;
    double mu = pars.muF;
    double chi = pars.chi;
    double m1 = MDIS->GetM1();
    parton p = pars.p;
    double yt = chi+(1.-chi)*t;
    double xt = chi/yt;
    double Fx  = pdf->xfx(chi,mu,p)/chi;
    double Fxt = pdf->xfx(xt,mu,p)/xt;
    //the plus prescription is defined between chi and 1
    //int_chi^1 C Q_1 dy/y
    //Q_1 = pdf, C = g[f]_+, f = 1/(1-z)_+, g = (1-z)*(shatp)/(8*shat)*1/N*fhat
    double result;
    result = (1-chi)*f_sub(yt,m1,Q)*(1./yt*Fxt-Fx)+Fx*(-chi*f_sub(chi*t,m1,Q));
    return result;
}

double MassiveDIS::q_sub(double x, const double muLog, const double muF, sfunction fun, parton part)
{
	sub_plus pars = {muLog, muF, x, this, fPDF, part};
	double res=0;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    res = integration::gauss(Sub_integrand,0.,1.,fPrec,NULL,&pars);
    gsl_set_error_handler (old_handler);
    return res;
}

/*
	Massless coefficient functions - taken from
	Furmanski, Petronzio Z. Phys. C - Particles and Fields, 11, 293-314 (1982)
 */

/*
	Plus prescription term
 */
double massless_f_plus(double x, double Q_over_muF)
{
	double result;
	result = ((1.+pow(x,2))/(1.-x)*(log((1.-x)/x)-3./4.+2*log(Q_over_muF))+1./4.*(9.+5.*x));
	return result;
}

/*
	Regular term
 */
double massless_f_regular(double x, sfunction fun)
{
	double result;
	switch(fun)
	{
		case F_1:
			result = -2.*x;
			break;

		case F_2:
			result = 0;
			break;

		case F_3:
			result = -(1.+x);
			break;

		default:
			cout << "Error. Abort." << endl;
			abort();
	}
	return result;
}

/*
	Integration
 */
struct massless_f_struct  {double Q, mu, x; PDFinterface::PDFs *pdf; parton p; sfunction fun;};
double F_massless_integrand(double t, void* par) 
{
	massless_f_struct pars = *(massless_f_struct *) par;
    PDFinterface::PDFs *pdf  = pars.pdf;
    double Q = pars.Q;
    double mu = pars.mu;
    double x = pars.x;
    parton p = pars.p;
    sfunction fun = pars.fun;
    double yt = x+(1.-x)*t;
    double xt = x/yt;
    double Fx  = pdf->xfx(x,mu,p)/x;
    double Fxt = pdf->xfx(xt,mu,p)/xt;
    //the plus prescription is defined between chi and 1
    //int_chi^1 C Q_1 dy/y
    //Q_1 = pdf, C = g[f]_+, f = 1/(1-z)_+, g = (1-z)*(shatp)/(8*shat)*1/N*fhat
    double result;
    result = (1.-x)*massless_f_plus(yt,Q/mu)*(1./yt*Fxt-Fx)-x*Fx*massless_f_plus(x*t,Q/mu)
    +(1.-x)/yt*massless_f_regular(yt,fun)*Fxt;
    return result;
}

double MassiveDIS::F_q_as_0(double x, const double Q, const double muF, sfunction fun, parton part) 
{
	massless_f_struct pars = {Q, muF, x, fPDF, part, fun};
	double res=0;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    res = integration::gauss(F_massless_integrand,0.,1.,fPrec,NULL,&pars);
    gsl_set_error_handler (old_handler);
    return res;
}

struct gluon_f_struct_0  {double muLog, muF, x; MassiveDIS *MDIS; PDFinterface::PDFs *pdf; parton p; sfunction fun;};
double F_gluon_sub_integrand(double t, void* par) 
{
	gluon_f_struct_0 pars = *(gluon_f_struct_0 *) par;
    PDFinterface::PDFs *pdf  = pars.pdf;
    double muF = pars.muF;
    double muLog = pars.muLog;
    MassiveDIS *MDIS = pars.MDIS;
    double xt = pars.x/t;
    double Fx  = pdf->xfx(xt,muF,g)/xt;
    double m1 = MDIS->GetM1();
    double m2 = MDIS->GetM2();
    sfunction fun = pars.fun;
    double result = (fun == F_3 ? -1. : 1.)*1./2*log(pow(muLog,2)/pow(m1,2))*1./t*1./2*(pow(t,2)+pow(1-t,2))*Fx+1./2*log(pow(muLog,2)/pow(m2,2))*1./t*1./2*(pow(t,2)+pow(1-t,2))*Fx;
    return result;
}

double MassiveDIS::g_sub(double x, const double muLog, const double muF, sfunction fun) 
{
	gluon_f_struct_0 pars = {muLog, muF, x, this, fPDF, g, fun};
	double res=0;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    res = integration::gauss(F_gluon_sub_integrand,x,1.,fPrec,NULL,&pars);
    gsl_set_error_handler (old_handler);
    return res;
}

double gluon_coeff(double t, double Q, MassiveDIS *MDIS, sfunction fun)
{
	double result;
	switch(fun)
	{
		case F_1:
			result = MDIS->gluon_f1(t,Q);
			break;

		case F_2:
			result = MDIS->gluon_f2(t,Q);
			break;

		case F_3:
			result = MDIS->gluon_f3(t,Q);
			break;

		default:
			cout << "Error. Abort." << endl;
			abort();
	}
	return result;
}

struct gluon_f_struct  {double Q, muF, x; sfunction fun; MassiveDIS *MDIS; PDFinterface::PDFs *pdf;};
double F_gluon_integrand(double t, void* par) 
{
	gluon_f_struct pars = *(gluon_f_struct *) par;
    PDFinterface::PDFs *pdf  = pars.pdf;
    double Q = pars.Q;
    double muF = pars.muF;
    MassiveDIS *MDIS = pars.MDIS;
    sfunction fun = pars.fun;
    double x = pars.x;
    double xt = x/t;
    double Ft  = pdf->xfx(t,muF,g)/t;
    double result = Ft*gluon_coeff(xt,Q,MDIS,fun);
    if(fun == F_1 || fun == F_3) result/=t; //different normalisation of the integral
    return result;
}

double MassiveDIS::Fgluon(double x, const double Q, const double muF, sfunction fun) 
{
	gluon_f_struct pars = {Q, muF, x, fun, this, fPDF};
	double res=0;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double ax = (1.+pow(fM1+fM2,2)/pow(Q,2))*x;
    if(ax>=1) return 0;
    res = integration::gauss(F_gluon_integrand,ax,1.,fPrec,NULL,&pars);
    gsl_set_error_handler (old_handler);
    return res;
}


double gluon_coeff_0(double t, double Q_over_muF, sfunction fun)
{
	double result;
	switch(fun)
	{
		case F_1:
			result = gluon_coeff_0(t,Q_over_muF,F_2)-1./2*(4.*t*(1.-t));
			break;

		case F_2:
			result = 1./2*((pow(1.-t,2)+pow(t,2))*(log((1.-t)/t)+2*log(Q_over_muF))-8.*pow(t,2)+8.*t-1);
			break;

		case F_3:
			result = 0; 
			break;

		default:
			cout << "Error. Abort." << endl;
			abort();
	}
	return result;
}

double F_gluon_integrand0(double t, void* par) 
{
	gluon_f_struct pars = *(gluon_f_struct *) par;
    PDFinterface::PDFs *pdf  = pars.pdf;
    double Q = pars.Q;
    double muF = pars.muF;
    sfunction fun = pars.fun;
    double x = pars.x;
    double xt = x/t;
    double Ft  = pdf->xfx(t,muF,g)/t;
    double result = gluon_coeff_0(xt,Q/muF,fun)*Ft*1./t;
    return result;
}

double MassiveDIS::Fgluon0(double x, const double Q, const double muF, sfunction fun) 
{
	gluon_f_struct pars = {Q, muF, x, fun, this, fPDF};
	double res=0;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    res = integration::gauss(F_gluon_integrand0,x,1.,fPrec,NULL,&pars);
    gsl_set_error_handler (old_handler);
    return res;
}


/*
	Gluon Contribution from Z. Phys. C 38 (err. ibid. C.39), Glück et al
 */

double MassiveDIS::gluon_f1(double x, const double Q)
{
	double result;
	double v      = gluon_v(x,Q);
	double vbar   = gluon_vbar(x,Q);
	double L      = gluon_L(x,Q);
	double Ltilde = gluon_Ltilde(x,Q);
	double m = fM1;
	double m2 = pow(fM1,2);
	double mp = fM2;
	double mp2 = pow(fM2,2);
	double Q2 = pow(Q,2);
	double Q4 = pow(Q,4);
	result = fQp/4.*(-v*vbar*(pow(1-2.*x,2)+fQm/fQp*m*mp/Q2*4.*x*(1.-x))
		+(1./2.-x*(1.-x)+(m2-mp2)/Q2*x*(1.-2.*x)+pow(m2-mp2,2)/Q4*pow(x,2)+fQm/fQp*m*mp/Q2*2.*x*(1.-x-x*(m2+mp2)/Q2))*L
		+(1./2.-x*(1.-x)+(mp2-m2)/Q2*x*(1.-2.*x)+pow(mp2-m2,2)/Q4*pow(x,2)+fQm/fQp*mp*m/Q2*2.*x*(1.-x-x*(mp2+m2)/Q2))*Ltilde
		);
	return result;
}


double MassiveDIS::gluon_f2(double x, const double Q)
{
	double result;
	double v      = gluon_v(x,Q);
	double vbar   = gluon_vbar(x,Q);
	double L      = gluon_L(x,Q);
	double Ltilde = gluon_Ltilde(x,Q);
	double m = fM1;
	double m2 = pow(fM1,2);
	double m4 = pow(fM1,4);
	double m6 = pow(fM1,6);
	double mp = fM2;
	double mp2 = pow(fM2,2);
	double mp4 = pow(fM2,4);
	double mp6 = pow(fM2,6);
	double Q2 = pow(Q,2);
	double Q4 = pow(Q,4);
	double Q6 = pow(Q,6);
	result = fQp*x*(
		v*vbar*(
			-1./2+4.*x*(1.-x)
			-((m2+mp2)/Q2-pow(m2-mp2,2)/Q4)*x*(1.-x))
		+1./4.*(
			1.-2.*x*(1.-x)+m2/Q2*(1+8*x-18*pow(x,2))
			+mp2/Q2*(1.-4.*x+6*pow(x,2))-fQm/fQp*2*m*mp/Q2
				-(m4+mp4)/Q4*2*x*(1.-3.*x)+m2*mp2/Q4*4.*x*(1.-5.*x)
				+(m6-m4*mp2-m2*mp4+mp6)/Q6*2.*pow(x,2))*L
		+1./4.*(
			1.-2.*x*(1.-x)+mp2/Q2*(1+8*x-18*pow(x,2))
			+m2/Q2*(1.-4.*x+6*pow(x,2))-fQm/fQp*2*mp*m/Q2
				-(mp4+m4)/Q4*2*x*(1.-3.*x)+mp2*m2/Q4*4.*x*(1.-5.*x)
				+(mp6-mp4*m2-mp2*m4+m6)/Q6*2.*pow(x,2))*Ltilde
		);
	return result;
}

double MassiveDIS::gluon_f3(double x, const double Q)
{
	double result;
	double v      = gluon_v(x,Q);
	double vbar   = gluon_vbar(x,Q);
	double L      = gluon_L(x,Q);
	double Ltilde = gluon_Ltilde(x,Q);
	double m2 = pow(fM1,2);
	double m4 = pow(fM1,4);
	double mp2 = pow(fM2,2);
	double mp4 = pow(fM2,4);
	double Q2 = pow(Q,2);
	double Q4 = pow(Q,4);
	result = (v*vbar*((m2-mp2)/Q2)*2.*x*(1.-x)
	         -(1./2-x*(1-x)+(m2-mp2)/Q2*x*(1.-2.*x)-(m4-mp4)/Q4*pow(x,2))*L
	         +(1./2-x*(1-x)+(mp2-m2)/Q2*x*(1.-2.*x)-(mp4-m4)/Q4*pow(x,2))*Ltilde
	         ); //+ or -?
	return result;
}

double MassiveDIS::gluon_L(double x, const double Q)
{
	double num = (1.+(pow(fM1,2)-pow(fM2,2))/pow(Q,2)*x/(1.-x)+gluon_v(x,Q)*gluon_vbar(x,Q));
	double den = (1.+(pow(fM1,2)-pow(fM2,2))/pow(Q,2)*x/(1.-x)-gluon_v(x,Q)*gluon_vbar(x,Q));
	return log(num/den);
}

double MassiveDIS::gluon_Ltilde(double x, const double Q)
{
	double num = (1.+(pow(fM2,2)-pow(fM1,2))/pow(Q,2)*x/(1.-x)+gluon_v(x,Q)*gluon_vbar(x,Q));
	double den = (1.+(pow(fM2,2)-pow(fM1,2))/pow(Q,2)*x/(1.-x)-gluon_v(x,Q)*gluon_vbar(x,Q));
	return log(num/den);
}

double MassiveDIS::gluon_v(double x, const double Q)
{
	return sqrt(1. - pow(fM1+fM2,2)/pow(Q,2)*x/(1.-x));
}

double MassiveDIS::gluon_vbar(double x, const double Q)
{
	return sqrt(1. - pow(fM1-fM2,2)/pow(Q,2)*x/(1.-x));
}

parton MassiveDIS::CCPartner(const parton p)
{

	parton cpart;

	if(p == u) cpart = dbar;
	else if(p == dbar) cpart = u;
	else if(p == d) cpart = ubar;
	else if(p == ubar) cpart = d;  

	else if(p == c) cpart = sbar;
	else if(p == sbar) cpart = c;
	else if(p == s) cpart = cbar;
	else if(p == cbar) cpart = s;


	if(p == b) cpart = tbar;
	else if(p == tbar) cpart = b;
	else if(p == t) cpart = bbar;
	else if(p == bbar) cpart = t;

	return cpart;

}

